import importlib
import inspect
import logging
import sys

import uvicorn
from slugify import slugify
from starlette.applications import Starlette
from starlette.types import ASGIApp

from .exceptions import NoServingModelsFoundError
from .pandas_serving import DataFrameModelServing
from .serving import (
    ModelServing,
    _index_endpoint,
    _kubernetes_liveness_endpoint,
    _kubernetes_readiness_endpoint,
)

EXCLUDED_SERVING_CLASSES = [ModelServing, DataFrameModelServing]
logger = logging.getLogger(__name__)


def compose_serving_models(module_name: str = "models", debug: bool = False) -> ASGIApp:
    try:
        python_module = importlib.import_module(module_name)
    except ModuleNotFoundError as exc:
        logger.critical(f"Cannot find Python module named {module_name}: {exc}")
        raise exc
    class_members = inspect.getmembers(sys.modules[module_name], inspect.isclass)
    serving_models = [
        class_
        for _, class_ in class_members
        if issubclass(class_, ModelServing) and class_ not in EXCLUDED_SERVING_CLASSES
    ]
    if not serving_models:
        raise NoServingModelsFoundError(f"Could not find any models in {python_module}")
    elif len(serving_models) == 1:
        model_serving = serving_models[0](debug=debug)
    else:
        model_serving = Starlette(debug=debug)
        for asgi_app in serving_models:
            model_serving.mount(f"/{slugify(asgi_app.__name__)}", asgi_app(debug=debug))
        model_serving.add_route("/", _index_endpoint, methods=["GET"])
        model_serving.add_route(
            "/liveness/", _kubernetes_liveness_endpoint, methods=["GET"]
        )
        model_serving.add_route(
            "/readiness/", _kubernetes_readiness_endpoint, methods=["GET"]
        )
    return model_serving


def run_model_serving(module_name: str = "models", debug: bool = False):
    asgi_app = compose_serving_models(module_name, debug)
    uvicorn.run(asgi_app, debug=debug)



import logging
from typing import Dict, Union

from starlette.exceptions import HTTPException

from .serving import ModelServing

logger = logging.getLogger(__name__)
try:
    import modin.pandas as pandas
except ImportError:
    try:
        import pandas
    except ImportError as e:
        logger.critical(
            f"Cannot import pandas. Please install foxcross using foxcross[pandas] or"
            f" foxcross[modin]"
        )
        raise e

try:
    import ujson as json
    from starlette.responses import UJSONResponse as JSONResponse
except ImportError:
    import json  # noqa: F401
    from starlette.responses import JSONResponse


class DataFrameModelServing(ModelServing):
    # TODO: probably should limit which choices are available for orient since to_dict and
    # to_json differ
    pandas_orient = "index"

    def predict(
        self, data: Union[pandas.DataFrame, Dict[str, pandas.DataFrame]]
    ) -> Union[pandas.DataFrame, Dict[str, pandas.DataFrame]]:
        """
        Method to define how the model performs a prediction.
        Must return a pandas DataFrame or a dictionary of pandas DataFrames
        """
        raise NotImplementedError(
            "You must implement your model serving's predict method"
        )

    def pre_process_input(
        self, data: Dict
    ) -> Union[pandas.DataFrame, Dict[str, pandas.DataFrame]]:
        pre_processed_data = super().pre_process_input(data)
        try:
            if data.get("multi_dataframe", None) is True:
                return {
                    key: pandas.read_json(value, orient=self.pandas_orient)
                    for key, value in pre_processed_data.items()
                    if key != "multi_dataframe"
                }
            else:
                return pandas.read_json(pre_processed_data, orient=self.pandas_orient)
        except (TypeError, KeyError) as exc:
            err_msg = f"Error reading in json: {exc}"
            logger.warning(err_msg)
            raise HTTPException(status_code=400, detail=err_msg)

    def post_process_results(
        self, data: Union[pandas.DataFrame, Dict[str, pandas.DataFrame]]
    ) -> JSONResponse:
        try:
            results = data.to_json(orient=self.pandas_orient)
        except AttributeError:
            results = {
                key: value.to_dict(orient=self.pandas_orient)
                for key, value in data.items()
            }
        return super().post_process_results(results)



import logging
import os
from pathlib import Path
from typing import Any, Dict, List

import aiofiles
from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.middleware.gzip import GZipMiddleware
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware
from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.templating import Jinja2Templates

from .enums import MediaTypes
from .exceptions import BadDataFormatError

try:
    import ujson as json
    from starlette.responses import UJSONResponse as JSONResponse
except ImportError:
    import json
    from starlette.responses import JSONResponse

SCRIPT_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
templates = Jinja2Templates(directory=str(SCRIPT_DIR / "templates"))
logger = logging.getLogger(__name__)


async def _index_endpoint(request: Request) -> Jinja2Templates.TemplateResponse:
    return templates.TemplateResponse(
        "index.html", {"request": request, "routes": request.app.routes}
    )


async def _kubernetes_liveness_endpoint(request: Request) -> PlainTextResponse:
    return PlainTextResponse("No problems")


async def _kubernetes_readiness_endpoint(request: Request) -> PlainTextResponse:
    return PlainTextResponse("No problems")


class ModelServing(Starlette):
    test_data_path = None

    def __init__(
        self, redirect_https: bool = False, gzip_response: bool = True, **kwargs
    ):
        try:
            test_data = Path(self.test_data_path)
        except TypeError as exc:
            logger.critical(
                f"Problem loading {self.test_data_path}: {exc}. Please make"
                f" sure you have defined your test_data_path on your model serving"
            )
            raise exc
        assert test_data.exists(), f"{self.test_data_path} does not exist"
        super().__init__(**kwargs)
        self.load_model()
        self.add_route("/", _index_endpoint, methods=["GET"])
        self.add_route("/liveness/", _kubernetes_liveness_endpoint, methods=["GET"])
        self.add_route("/readiness/", _kubernetes_readiness_endpoint, methods=["GET"])
        self.add_route("/predict/", self._predict_endpoint, methods=["HEAD", "POST"])
        self.add_route("/predict-test/", self._predict_test_endpoint, methods=["GET"])
        self.add_route("/input-format/", self._input_format_endpoint, methods=["GET"])
        if gzip_response is True:
            self.add_middleware(GZipMiddleware)
        if redirect_https is True:
            self.add_middleware(HTTPSRedirectMiddleware)
        self._media_types = [
            MediaTypes.ANY.value,
            MediaTypes.ANY_APP.value,
            MediaTypes.JSON.value,
        ]

    def load_model(self):
        """Hook to load a model or models"""
        pass

    def predict(self, data: Any) -> Any:
        """
        Method to define how the model performs a prediction.
        Must return JSON serializable data
        """
        raise NotImplementedError(
            "You must implement your model serving's predict method"
        )

    async def _read_test_data(self) -> Dict:
        async with aiofiles.open(self.test_data_path, mode="rb") as f:
            contents = await f.read()
        try:
            return json.loads(contents.decode("utf-8"))
        except TypeError as exc:
            raise HTTPException(
                status_code=500, detail=f"Failed to read test data: {exc}"
            )

    async def _predict_endpoint(self, request: Request) -> JSONResponse:
        if request.method == "HEAD":
            return JSONResponse()
        self._validate_http_headers(request, "content-type", self._media_types, 415)
        self._validate_http_headers(request, "accept", self._media_types, 406)
        json_data = await request.json()
        pre_processed_input = self.pre_process_input(json_data)
        try:
            results = self.predict(pre_processed_input)
        except BadDataFormatError as exc:
            logger.warning(f"Bad data format inputted to the predict endpoint: {exc}")
            raise HTTPException(status_code=400, detail=str(exc))
        return self.post_process_results(results)

    async def _predict_test_endpoint(self, request: Request) -> JSONResponse:
        self._validate_http_headers(request, "accept", self._media_types, 406)
        test_data = await self._read_test_data()
        pre_processed_input = self.pre_process_input(test_data)
        results = self.predict(pre_processed_input)
        return self.post_process_results(results)

    async def _input_format_endpoint(self, request: Request) -> JSONResponse:
        self._validate_http_headers(request, "accept", self._media_types, 406)
        test_data = await self._read_test_data()
        pre_processed_input = self.pre_process_input(test_data)
        return self.post_process_results(pre_processed_input)

    @staticmethod
    def _validate_http_headers(
        request: Request, header: str, media_types: List[str], invalid_status_code: int
    ):
        if not request.headers.get(header):
            err_msg = (
                f"Missing http header {header}. Please provide one with an appropriate"
                f" media type. Possible types are {', '.join(media_types)}"
            )
            logger.warning(err_msg)
            raise HTTPException(status_code=400, detail=err_msg)
        elif not any(x in request.headers[header] for x in media_types):
            err_msg = (
                f"Media types {request.headers[header]} in {header} header are not"
                f" supported. Supported types are {', '.join(media_types)}"
            )
            logger.warning(err_msg)
            raise HTTPException(status_code=invalid_status_code, detail=err_msg)

    @staticmethod
    def _get_response(data: Any) -> JSONResponse:
        try:
            return JSONResponse(data)
        except TypeError as exc:
            raise HTTPException(
                status_code=500,
                detail=f"Error trying to serialize response data to JSON: {exc}",
            )

    def pre_process_input(self, data: Any) -> Any:
        """Hook to enable pre-processing of input data"""
        return data

    def post_process_results(self, data: Any) -> JSONResponse:
        """Hook to enable post-processing of output data"""
        return self._get_response(data)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['foxcross']

package_data = \
{'': ['*'], 'foxcross': ['templates/*']}

install_requires = \
['aiofiles>=0.4.0,<0.5.0',
 'jinja2>=2.10,<3.0',
 'python-slugify[unidecode]>=3.0,<4.0',
 'starlette>=0.12.0,<0.13.0',
 'uvicorn>=0.7.1,<0.8.0']

extras_require = \
{'modin': ['modin>=0.5.1,<0.6.0'],
 'pandas': ['pandas>=0.24.2,<0.25.0'],
 'performance': ['ujson>=1.35,<2.0', 'modin>=0.5.1,<0.6.0'],
 'ujson': ['ujson>=1.35,<2.0']}

setup_kwargs = {
    'name': 'foxcross',
    'version': '0.3.0',
    'description': 'AsyncIO serving for data science models',
    'long_description': '## Foxcross\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/python/black)\n[![License](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://github.com/laactech/foxcross/blob/master/LICENSE.md)\n[![Build Status](https://travis-ci.org/laactech/foxcross.svg?branch=master)](https://travis-ci.org/laactech/foxcross)\n[![Build status](https://ci.appveyor.com/api/projects/status/ufbm8hrkp4whol5a?svg=true)](https://ci.appveyor.com/project/laactech/foxcross)\n\nAsyncIO serving for data science models built on [Starlette](https://www.starlette.io/)\n\n**Documentation**: https://www.foxcross.dev/\n\n**Requirements**: Python 3.6+\n\n## Quick Start\nInstallation using `pip`:\n```bash\npip install foxcross\n```\n\nCreate some test data and a simple model in the same directory to be served:\n\n`data.json`\n```json\n[1,2,3,4,5]\n```\n\n`models.py`\n```python\nfrom foxcross.runner import run_model_serving\nfrom foxcross.serving import ModelServing\n\nclass AddOneModel(ModelServing):\n    test_data_path = "data.json"\n    \n    def predict(self, data):\n        return [x + 1 for x in data]\n\nif __name__ == "__main__":\n    run_model_serving()\n```\n\nRun the model locally:\n```bash\npython models.py\n```\n\nNavigate to `localhost:8000/predict-test/`, and you should see the list incremented by 1.\nYou can visit `localhost:8000/` to see all the available routes for your model.\n',
    'author': 'Steven Pate',
    'author_email': 'steven@laac.io',
    'url': 'https://github.com/laactech/foxcross',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
